{-# LANGUAGE KindSignatures #-}
{-# LANGUAGE GeneralizedNewtypeDeriving #-}
{-# LANGUAGE DataKinds #-}
module SlideRule (
  SlideRule,
  Scale(..),
  ScalePosition(..),
  invertScale,

  mark,
  align,
  turn,

  readCursor,
  readAt,

  runSlideRule,
  printAnswer,
  
) where

import Control.Applicative
import Control.Monad.Trans.State
import Data.Fixed (mod')
import Numeric

newtype SlideRule a = SlideRule (State SlideState a)
  deriving (Monad, Applicative, Functor)

data SlideState = SlideState
  { slideWheel  :: Double
  , slideCursor :: Double
  }

initialState :: SlideState
initialState = SlideState
  { slideWheel = 0
  , slideCursor = 0
  }

runSlideRule :: SlideRule a -> a
runSlideRule (SlideRule m) = evalState m initialState

printAnswer :: SlideRule Double -> String
printAnswer s = showFFloat (Just 3) (runSlideRule s) ""

instance Show a => Show (SlideRule a) where
  show = show . runSlideRule

data ScalePosition = OnWheel | OffWheel

class IsOnWheel (pos :: ScalePosition) where
  scaleWheel :: scale pos -> Bool

instance IsOnWheel OnWheel where
  scaleWheel _ = True

instance IsOnWheel OffWheel where
  scaleWheel _ = False

data Scale (w :: ScalePosition) = Scale
  { scaleName :: String
  , scaleFun  :: Double -> Double -- ^ scale domain -> position
  , scaleFun' :: Double -> Double -- ^ position -> scale domain
  }

invertScale :: String -> Scale x -> Scale y
invertScale n s = Scale
  { scaleName = n
  , scaleFun  = \x -> 1 - scaleFun s x -- mod' _ 1?
  , scaleFun' = \x -> scaleFun' s (1-x)
  }

getCursorPos :: SlideRule Double
getCursorPos = SlideRule (gets slideCursor)

getWheelPos :: SlideRule Double
getWheelPos = SlideRule (gets slideWheel)

setCursorPos :: Double -> SlideRule ()
setCursorPos x = SlideRule (modify (\s -> s { slideCursor = mod' x 1 }))

setWheelPos :: Double -> SlideRule ()
setWheelPos x = SlideRule (modify (\s -> s { slideWheel = mod' x 1 }))

readCursor :: IsOnWheel pos => Scale pos -> SlideRule Double
readCursor scale = do
  c <- getCursorPos
  w <- if scaleWheel scale then getWheelPos else return 0
  let c' = mod' (c-w) 1
  return (scaleFun' scale c')

readAt :: (IsOnWheel a, IsOnWheel b) => Scale a -> Double -> Scale b -> SlideRule Double
readAt refScale refValue outScale =
 do x <- correction (scaleFun refScale refValue)
    pure (scaleFun' outScale x)
  where
    correction x
      | scaleWheel refScale, not (scaleWheel outScale) =
        do p <- getWheelPos
           pure (mod' (x+p) 1)
      | not (scaleWheel refScale), scaleWheel outScale =
        do p <- getWheelPos
           pure (mod' (x-p) 1)
      | otherwise = pure x

-- | Set the cursor to the value on the given scale
mark :: IsOnWheel pos => Scale pos -> Double -> SlideRule ()
mark scale value

  | scaleWheel scale = do w <- getWheelPos
                          setCursorPos (w + frac)

  | otherwise        = setCursorPos frac
  where
  frac = scaleFun scale value

-- | Turn the wheel to align a value on the wheel with a value off the wheel.
align :: Scale OffWheel -> Double -> Scale OnWheel -> Double -> SlideRule ()
align offScale offValue onScale onValue =
  setWheelPos (scaleFun offScale offValue - scaleFun onScale onValue)

-- | Turn the wheel so that a scale on the wheel has a specific value
-- on the cursor.
turn :: Scale OnWheel -> Double -> SlideRule ()
turn scale value =
 do c <- getCursorPos
    let v = scaleFun scale value
    setWheelPos (c - v)

resetWheel :: SlideRule ()
resetWheel = setWheelPos 0

{-
multiply :: Double -> Double -> String
multiply x y = printAnswer $
  do mark dScale  x
     move  ciScale y
     mark ciScale 1
     return dScale

fourthPowCIK :: Double -> String
fourthPowCIK x = printAnswer $
  do mark kScale x
     move ciScale x
     mark ciScale 1
     return kScale

fourthPowCCIA :: Double -> String
fourthPowCCIA x = printAnswer $
  do move ciScale x
     mark cScale x
     return aScale
-}
